'******************************************************************************
' Copyright (c) 2013 Dave Hein
' See end of file for terms of use.
'******************************************************************************
CON
  _clkmode = xinput
  _xinfreq = $4e495053 ' SPIN

  DOUBLE_QUOTE = $22

DAT
  printflag long 0
  globalflag long 0

OBJ
  c : "clibsd"
  sys : "sysdefs"
  vs : "varsubs"

PUB main(argc, argv) | ptr, argi

  c.enter

  ' Check option flags
  argi := 1
  repeat while argi < argc and byte[long[argv][argi]] == "-"
    if strcomp(long[argv][argi], string("-p"))
      printflag := 1
    elseif strcomp(long[argv][argi], string("-x"))
      globalflag := 1
    else
      usage
    argi++

  ' Print if no other parms
  if argi == argc
    if globalflag
      PrintVars(1)
    elseif printflag
      PrintVars(2)
    else
      PrintVars(0)
    c.exit(0)

  ' Print if -p was specified
  if printflag
    repeat while argi < argc
      ptr := vs.FindVar(long[argv][argi])
      ifnot ptr
        c.printf1(string("%s not found\n"), long[argv][argi])
      else
        PrintVar(ptr, 2)
      argi++
    c.exit(0)

  repeat while argi < argc
    ProcessParm(long[argv][argi])
    argi++

  c.exit(0)

PUB ProcessParm(name) | ptr, len1, len2, space, value, noequal

  ' Check for "=" and get value
  value := vs.FindChar(name, "=")
  ifnot byte[value]
    noequal := 1
  else
    noequal := 0
    byte[value++] := 0

  ' Check if variable already exists
  if (ptr := vs.FindVar(name))
    if globalflag and noequal
      byte[ptr] := "0"
      return
    if byte[ptr] == "0"
      globalflag := 1
    ptr := vs.RemoveEntry(ptr)
  else
    ptr := vs.FindEnd

  ' Check if space is available
  space := sys#environ_vars_end - ptr + 1
  len1 := strsize(name)
  len2 := strsize(value)
  if space - len1 - len2 - 4 < 50
    c.printf2(string("space = %d, need = %d\n"), space, len1 + len2 + 4)
  if (space < len1 + len2 + 4)
    c.printf0(string("Not enough variable space\n"))
    c.exit(1)

  ' Add variable
  if globalflag
    byte[ptr++] := VS#GLOBAL_FLAG
  else
    byte[ptr++] := VS#LOCAL_FLAG
  c.strcpy(ptr, name)
  ptr += len1 + 1
  c.strcpy(ptr, value)
  ptr += len2 + 1
  byte[ptr] := 0

PUB usage
  c.printf0(string("usage: declare [-x] [-p] [var[=[value]]...]\n"))
  c.exit(1)

PUB PrintVars(mode) | ptr
  ptr := sys#environ_vars
  repeat while byte[ptr]
    ptr := PrintVar(ptr, mode)

PUB PrintVar(ptr, mode) | vartype, name, value
  vartype := byte[ptr++]
  name := ptr
  ptr += strsize(ptr) + 1
  value := ptr
  ptr += strsize(ptr) + 1
  if mode
    if vartype == VS#GLOBAL_FLAG
      c.printf2(string("declare -x %s=", DOUBLE_QUOTE, "%s", DOUBLE_QUOTE, "\n"), name, value)
    elseif mode == 2 and vartype == VS#LOCAL_FLAG
      c.printf2(string("declare -- %s=", DOUBLE_QUOTE, "%s", DOUBLE_QUOTE, "\n"), name, value)
  elseif vartype == VS#GLOBAL_FLAG or vartype == VS#LOCAL_FLAG
    if ContainsBlank(value)
      c.printf2(string("%s=", DOUBLE_QUOTE, "%s", DOUBLE_QUOTE, "\n"), name, value)
    else
      c.printf2(string("%s=%s\n"), name, value)
  return ptr

PUB ContainsBlank(ptr)
  repeat while byte[ptr]
    if byte[ptr++] == " "
      return 1

{{
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
}}
