'******************************************************************************
' Copyright (c) 2011, 2012, 2013, Dave Hein
' See end of file for terms of use.
'******************************************************************************
CON
  _clkmode = xinput
  _xinfreq = $4e495053 ' SPIN

  CHECKWORD = $dead1eaf
  NEW_LINE = 10
  eeprom_pin = 28

OBJ
  c      : "clibsd"
  mem    : "cmalloc"
  ser    : "cserial"
  eeprom : "pasm_i2c_driver"
  fsrwx  : "fsrwx"
  sys    : "sysdefs"
  vs     : "varsubs"
  g      : "glob"

DAT
  badflag long 0
  buffer byte 0[100]
  strbuf byte 0[100]
  i2c_cog long 0
  nullstr long 0
  
DAT
  tokens long 0[20]
  commands byte
  byte "dumpm", 0, "peekm", 0, "pokem", 0
  byte "stack", 0, "malloc", 0, "help", 0, "exit", 0
  byte "dump", 0, "peek", 0, "poke", 0, "cd", 0, "echo", 0, "pwd", 0, 0
  i2cmsg byte "The I2C driver is not loaded.  Run loadi2c\n", 0

PUB main(argc, argv) | i, blocklongs, num, index, outfile, infile, startaddr, endaddr, TopAddr, TopObjList
  c.enter
  SetShell(string("/bin/diag"))
  i2c_cog := eeprom.Initialize(eeprom_pin)
  'c.printf1(string("long[sys#i2c_cog] = %8.8x\n"), long[sys#i2c_cog])
  REPEAT
    c.CloseRedirect

    c.printf0(STRING("diag> "))
    c.gets(@buffer)
    argc := tokenize(@buffer, @tokens)
    argc := Redirection(argc, @tokens)

    IF (argc == 0)
      NEXT

    num := 0
    index := @commands
    repeat while byte[index]
      if c.strcmp(tokens[0], index) == 0
        quit
      index += strsize(index) + 1
      num++

    case num
      0 :
        c.sscanf1(tokens[1], STRING("%x"), @index)
        c.sscanf1(tokens[2], STRING("%x"), @num)
        DumpMemory(index, num)
      1 :
        c.sscanf1(tokens[1], STRING("%x"), @index)
        c.printf2(STRING("Memory[%04x] = %02x\n"), index, byte[index])
      2 :
        c.sscanf1(tokens[1], STRING("%x"), @index)
        c.sscanf1(tokens[2], STRING("%x"), @num)
        byte[index] := num
      3:
        c.CheckStackSpace '(argc, @tokens)
      4:
        c.CheckMallocSpace
      5:
        if (infile := c.fopen(string("/manpages/diaghelp.txt"), string("r")))
          repeat while c.fgets(@buffer, 100, infile)
            c.fputs(@buffer, c.getstdout)
          c.fclose(infile)
      6:
        exit(0)
      7 :
        ifnot i2c_cog
          c.printf0(@i2cmsg)
          next
        c.sscanf1(tokens[1], STRING("%x"), @index)
        c.sscanf1(tokens[2], STRING("%x"), @num)
        repeat i from 0 to num - 1
          if ((i & 15) == 0)
            c.printf1(STRING("\n%05x:"), i + index)
          c.printf1(STRING(" %02x"), eeprom.ReadByte(28, $A0, i + index))
        c.printf0(STRING("\n"))
      8 :
        ifnot i2c_cog
          c.printf0(@i2cmsg)
          next
        c.sscanf1(tokens[1], STRING("%x"), @index)
        c.printf2(STRING("EEPROM[%04x] = %02x\n"), index, eeprom.ReadByte(28, $A0, index))
      9 :
        ifnot i2c_cog
          c.printf0(@i2cmsg)
          next
        c.sscanf1(tokens[1], STRING("%x"), @index)
        c.sscanf1(tokens[2], STRING("%x"), @num)
        eeprom.WriteByte(28, $A0, index, num)
        repeat while eeprom.WriteWait(28, $A0, index)
      10 :
        ChangeDirectory(argc, @tokens)
      11 :
        Echo(argc, @tokens)
      12 :
        PrintWD(argc, @tokens)
      other:
        Execute(argc, @tokens)

  exit(0)

PUB SetShell(name) | infile, handle
  infile := c.fopen(name, string("r"))
  if (infile)
    handle := long[infile][c#stream_handle]
    long[sys#shell_size] := fsrwx.hget_filesize(handle)
    long[sys#shell_sector] := fsrwx.hget_first_sector(handle)
    c.fclose(infile)

PUB exit(val)
  SetShell(string("/bin/shell"))
  cogstop(i2c_cog-1)
  c.exit(val)

PUB Execute(argc, argv) | target[20]
  'c.printf0(string("Execute\n"))
  'PrintArgv(argc, argv)
  'c.printf1(string("Trace %d\n"), 9)
  result := FindExecTarget(long[argv], @target)
  'c.printf1(string("Trace %d\n"), 20)
  if result => 2 and result =< 4
    long[argv] := @target
    if result == 4
      c.printf0(string("Can't execute stand-alone files\n"))
      return
    c.run(argc, argv, result - 2)
    c.printf1(string("Could not run %s\n"), long[argv])
  elseif result == 1
    c.printf0(string("Can't execute script files\n"))

PUB SkipChar(ptr, val)
  repeat while byte[ptr]
    if byte[ptr] <> val
      quit
    ptr++
  return ptr

PUB FindChar(ptr, val)
  repeat while byte[ptr]
    if byte[ptr] == val
      quit
    ptr++
  return ptr

PUB PrintWD(argc, argv)
  if c.OpenRedirect '(argc, argv)
    return 1
  c.getcwd(@buffer, 100)
  c.printf1(string("%s\n"), @buffer)

PUB ChangeDirectory(argc, argv) | path[25]
  if c.OpenRedirect '(argc, argv)
    return 1
  if argc > 2
    c.printf0(string("cd [path]\n"))
    return 1
  if argc == 1
    c.strcpy(@path, string("/"))
  else
    c.resolve_path(long[argv][1], @path)
    ifnot strcomp(@path, string("/"))
      ifnot c.getmod(@path) & %10000
        c.printf1(string("%s is not a directory\n"), @path)
        return 1
  if c.chdir(@path)
    c.printf1(string("Could not find path %s\n"), @path)
    return 2
  vs.SaveVar(string("PWD"), @path, vs#GLOBAL_FLAG)

PUB Echo(argc, argv) | i, ptr, space, globdata, str[25]
  if c.OpenRedirect '(argc, argv)
    return 1
  if argc > 1
    space := 0
    repeat i from 1 to argc - 1
      globdata := 0
      repeat while (ptr := g.glob(long[argv][i], @globdata, @str))
        c.printf2(string("%s%s"), @space, ptr)
        space := " "
  c.printf0(string("\n"))

PUB FindExecTarget(fname, pathstr) | temp, infile, typestr[8]
  'c.printf1(string("FindExecTarget %s\n"), fname)
  c.strcpy(pathstr, fname)
  'c.printf1(string("pathstr = %s\n"), pathstr)
  'c.CheckStackSpace
  'c.CheckMallocSpace
  'c.printf1(string("Trace %d\n"), 10)
  repeat 2
    'c.printf1(string("Trace %d\n"), 11)
    'c.printf0(string("repeat\n"))
    if (infile := c.fopen(pathstr, string("r")))
      'c.printf1(string("infile nonzero -- %d\n"), infile)
      quit
    'c.printf1(string("infile zero -- %d\n"), infile)
    temp := FindChar(pathstr, "/")
    if byte[temp]
      quit
    c.strcpy(pathstr, string("/bin/"))
    c.strcat(pathstr, fname)

  'c.printf1(string("infile equals %d\n"), infile)
  'c.printf1(string("Trace %d\n"), 12)

  ifnot infile
    c.printf1(string("1Could not open %s\n"), fname)
    return

  longfill(@typestr, 0, 8)
  c.fread(@typestr, 1, 32, infile)
  c.fclose(infile)

  'c.printf1(string("Trace %d\n"), 13)
  if typestr == $4e495053 ' SPIN
    result := 2
  elseif typestr == $50504143 ' CAPP
    result := 3
  elseif typestr => 80_000_000 and typestr =< 120_000_000 ' Stand-alone
    result := 4
  elseifnot c.strncmp(@typestr, string("#shell"), 6)
    'c.printf1(string("%s is a shell script\n"), fname)
    result := 1
  else
    c.printf1(string("%s is not an executable file\n"), fname)

PUB Redirection(argc, argv) | i, argc1, ptr, char, outptr, inptr, outstr
  if argc =< 0
    return argc
  i := 1
  argc1 := 1
  inptr := @nullstr
  outptr := @nullstr
  repeat while i < argc
    ptr := long[argv][i]
    char := byte[ptr++]
    if char == "<"
      if byte[ptr]
        inptr := ptr
      elseif ++i < argc
        inptr := long[argv][i]
      vs.SaveVar(string("<"), inptr, vs#LOCAL_FLAG)
    elseif char == ">"
      if byte[ptr] == ">"
        ptr++
        outstr := STRING(">>")
      else
        outstr := STRING(">")
      if byte[ptr]
        outptr := ptr
      elseif ++i < argc
        outptr := long[argv][i]
      vs.SaveVar(outstr, outptr, vs#LOCAL_FLAG)
    elseif char
      long[argv][argc1++] := long[argv][i]
    i++
  'long[argv][argc1] := inptr
  'long[argv][argc1+1] := outptr
  return argc1

' This routine extracts tokens from the input string that
' are seperated by at least one space character
PUB tokenize(ptr, argv) | num
  num := 0
  repeat
    ptr := SkipChar(ptr, " ")
    ifnot byte[ptr]
      quit
    num++
    if byte[ptr] == $22
      long[argv] := ++ptr
      ptr := FindChar(ptr, $22)
    else
      long[argv] := ptr
      ptr := FindChar(ptr, " ")
    argv += 4
    IF (BYTE[ptr] == 0)
      QUIT
    BYTE[ptr++] := 0
  RETURN num
  
PUB DumpMemory(address, num) | i, ptr
  ptr := address
  i := 0
  REPEAT WHILE (i < num)
    IF ((i & 15) == 0)
      c.printf1(STRING("\n%05x:"), i + address)
    c.printf1(STRING(" %02x"), BYTE[ptr][i])
    i++
  c.printf0(STRING("\n"))
{
PUB CheckStackSpace(argc, argv1) | addr, laststackaddr, entryptr, cognum, AppAddr, dbase
  cognum := cogid
    
  dbase := word[10]
  
  laststackaddr := dbase + 800 - 4
  repeat addr from laststackaddr to 0 step 4
    if long[addr] <> sys#CHECKWORD
      quit
    result++
  c.printf1(string("The last %d longs have not been used on the stack\n"), result)
  
PUB CheckMallocSpace | ptr, numalloc, numfree
  numalloc := 0
  numfree := 0
  ptr := long[sys#malloclist]
  c.printf0(string("MallocList\n"))
  repeat while(ptr)
    c.printf2(string("addr = %04x size = %d\n"), ptr, word[ptr][1])
    numalloc += word[ptr][1]
    ptr := word[ptr]
  ptr := long[sys#memfreelist]
  c.printf0(string("FreeList\n"))
  repeat while(ptr)
    c.printf2(string("addr = %04x size = %d\n"), ptr, word[ptr][1])
    numfree += word[ptr][1]
    ptr := word[ptr]
  c.printf2(string("%d bytes allocated, %d bytes free\n"), numalloc, numfree)
}
{{
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
}}
