'******************************************************************************
' Copyright (c) 2011, 2012, Dave Hein
' See end of file for terms of use.
'******************************************************************************
CON
  _clkmode = xinput
  _xinfreq = $4e495053 ' SPIN

OBJ
  c : "clibsd"
  sm : "smalloc"

CON
  s_next = 0          ' Word index for the linked-list next pointer
  s_str  = 2          ' Word index for the string data
  NUMLINES = 1000000  ' Set NUMLINES to a large value so it doesn't limit
  BUFFER_SIZE = 2000  ' Define a buffer size of 2000 bytes

VAR
  long buffer[BUFFER_SIZE / 4] ' Ensure that the buffer is long-aligned
  byte linebuf[200]            ' This is the input line buffer

DAT
  ' This flag indicates if there is a line in linebuf -- 1 for file1 and 2 for file2
  pending long 0

  ' List 1
  head1 long 0
  tail1 long 0
  level1 long 0
  eofflag1 long 0

  ' List 2
  head2 long 0
  tail2 long 0
  level2 long 0
  eofflag2 long 0

' This program implements the standard UNIX diff function between two files  
PUB main(argc, argv) | retval1, retval2, infile1, infile2, line1, line2, num1, num2, dashflag, next_1
  c.enter
  line1 := 0
  line2 := 0
  if (argc <> 3)
    usage
  ' Open file 1
  if (not(infile1 := c.fopen(long[argv][1], string("r"))))
    c.printf1(string("Could not open %s\n"), word[argv][1])
    c.exit(1)
  ' Open file 2
  if (not(infile2 := c.fopen(long[argv][2], string("r"))))
    c.printf1(string("Could not open %s\n"), word[argv][2])
    c.fclose(infile1)
    c.exit(2)
  ' Initialize the line buffer
  sm.smallocinit(@buffer, BUFFER_SIZE)
  ' Begin main loop
  repeat
    ' Read lines into list1 and list2
    FillBuffers(infile1, infile2)
    ' Check if list1 and list2 are empty
    if (not head1 and not head2)
      quit
    ' Compare the top line on the two lists
    if (not head1 or not head2 or c.strcmp((@word[head1][s_str]),(@word[head2][s_str])))
      ' Find matching lines if the top lines do not match
      FindMatch(@num1, @num2)
      ' Handle the case where the top line on list2 matches a later line on list1
      if (num1 > 0 and num2 == 0)
        dashflag := 0
        c.printf1(string("%d"), line1 + 1)
        if (num1 > 1)
          c.printf1(string(",%d"), line1 + num1)
        c.printf1(string("d%d\n"), line2)
      ' Handle the case where the top line on list1 matches a later line on list2
      elseif (num2 > 0 and num1 == 0)
        dashflag := 0
        c.printf2(string("%dd%d"), line1, line2 + 1)
        if (num2 > 1)
          c.printf1(string(",%d"), line2 + num2)
        c.printf0(string("\n"))
      ' Otherwise, handle the case where later lines on list1 and list2 match
      else
        dashflag := 1
        c.printf1(string("%d"), line1 + 1)
        if (num1 > 1)
          c.printf1(string(",%d"), line1 + num1)
        c.printf1(string("c%d"), line2 + 1)
        if (num2 > 1)
          c.printf1(string(",%d"), line2 + num2)
        c.printf0(string("\n"))
      ' Print lines from list1 and free them
      if (num1 > 0)
        line1 += num1
        repeat while (num1-- > 0)
          c.printf1(string("< %s"),(@word[head1][s_str]))
          next_1 := word[head1]{s_next}
          sm.sfree(head1)
          head1 := next_1
          level1--
      ' Print dashes if we are printing lines from both lists
      if (dashflag)
        c.printf0(string("---\n"))
      ' Print lines from list2 and free them
      if (num2 > 0)
        line2 += num2
        repeat while (num2-- > 0)
          c.printf1(string("> %s"),(@word[head2][s_str]))
          next_1 := word[head2]{s_next}
          sm.sfree(head2)
          head2 := next_1
          level2--
    ' Otherwise, don't print the top lines
    else
      ' Free the top line from list1
      if (head1)
        line1++
        next_1 := word[head1]{s_next}
        sm.sfree(head1)
        head1 := next_1
        level1--
      ' Free the top line from list2
      if (head2)
        line2++
        next_1 := word[head2]{s_next}
        sm.sfree(head2)
        head2 := next_1
        level2--
    ' Clear the tails if lists are empty
    if (not head1)
      tail1 := 0
    if (not head2)
      tail2 := 0
  ' Close the files before exiting
  c.fclose(infile1)
  c.fclose(infile2)
  c.exit(0)

' Print out the usage statement and exit      
PUB usage
  c.printf0(string("usage: diff file1 file2\n"))
  c.exit(1)

' This routine reads lines from the input files and adds them
' to list1 and list2 until one of the following occurs:
' - The end of file is reached on both files
' - There are NUMLINES in the buffer for both files
' - There is no more room left in the buffer to add a line
PUB FillBuffers(infile1, infile2) | next_1
  repeat
    ' Quit if EOF or there are NUMLINES on the list for both files
    if ((eofflag1 or level1 => NUMLINES) and(eofflag2 or level2 => NUMLINES))
      quit
    ' Read a line from an input file if one is not pending
    if (not pending)
      ' Read from file 1 if list1 has fewer lines than list2
      if (not eofflag1 and (level1 =< level2 or eofflag2))
        if (c.fgets(@linebuf, 200, infile1) =< 0)
          eofflag1 := 1
          next
        pending := 1
      ' Otherwise, read from file 2
      elseif (not eofflag2)
        if (c.fgets(@linebuf, 200, infile2) =< 0)
          eofflag2 := 1
          next
        pending := 2
    ' Quit if no lines are pending - EOF on both files
    if (not pending)
      quit
    ' Attempt to allocate a line buffer, and quit if there is no room
    next_1 := sm.smalloc(strsize(@linebuf) + 1)
    if (not next_1)
      quit
    c.strcpy((@word[next_1][s_str]), @linebuf)
    ' Add the line to list1
    if (pending == 1)
      if (tail1)
        word[tail1]{s_next} := next_1
      else
        head1 := next_1
      tail1 := next_1
      level1++
    ' Add the line to list2
    elseif (pending == 2)
      if (tail2)
        word[tail2]{s_next} := next_1
      else
        head2 := next_1
      tail2 := next_1
      level2++
    pending := 0

' This routine searches for a line that matches in list1 and list2.  It returns
' the smallest number of lines between list1 and list2 where a match is found.
PUB FindMatch(pnum1, pnum2) | minsum, num1, minnum1, count1, num2, minnum2, count2, next1, next2
  minsum := 1000000
  next1 := head1
  next2 := head2
  minnum1 := level1
  minnum2 := level2
  num1 := 0
  count1 := level1
  repeat while (count1 > 0)
    num2 := 0
    count2 := level2
    next2 := head2
    repeat while (count2 > 0)
      if (c.strcmp((@word[next1][s_str]),(@word[next2][s_str])) == 0)
        if (num2 + num1 < minsum)
          minnum2 := num2
          minnum1 := num1
          minsum := num1 + num2
          quit
      next2 := word[next2]{s_next}
      count2--
      num2++
    next1 := word[next1]{s_next}
    count1--
    num1++
  long[pnum1] := minnum1
  long[pnum2] := minnum2

{{
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
}}
