CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  eeprom_pin = 28

OBJ
  c   : "clibsd"
  sys : "sysdefs"
  ut  : "unixtime"
  vs  : "varsubs"
  spi : "safe_spi_p"
  ser : "cserial"
  fsrwx : "fsrwx"

PUB WritePage(SCL, devSel, addrReg, dataPtr, count) : ackbit
  ' Write three address bytes
  devSel |= $A0 | ((addr >> 15) & %1110)
  Start(SCL)
  Write(SCL, devSel)
  Write(SCL, addr >> 8 & $FF)
  Write(SCL, addr & $FF)
  ' Write four data bytes          
  repeat 4
    Write(SCL, byte[addr++])
  Stop(SCL)
  ' Wait till the write is done
  repeat
    Start(SCL)
    ackbit := Write(SCL, devSel)
    Stop(SCL)
  while ackbit
xxxxx

PUB Initialize(SCL) | SDA           ' An I2C device may be left in an
  SDA := SCL + 1                    '  invalid state and may need to be
  outa[SCL] := 1                    '   reinitialized.  Drive SCL high.
  dira[SCL] := 1
  dira[SDA] := 0                    ' Set SDA as input
  repeat 9
    outa[SCL] := 0                  ' Put out up to 9 clock pulses
    outa[SCL] := 1
    if ina[SDA]                     ' Repeat if SDA not driven high
      quit                          '  by the EEPROM

PUB Start(SCL) | SDA                ' SDA goes HIGH to LOW with SCL HIGH
  SDA := SCL + 1
  outa[SCL]~~                       ' Initially drive SCL HIGH
  dira[SCL]~~
  outa[SDA]~~                       ' Initially drive SDA HIGH
  dira[SDA]~~
  outa[SDA]~                        ' Now drive SDA LOW
  outa[SCL]~                        ' Leave SCL LOW
  
PUB Stop(SCL) | SDA                 ' SDA goes LOW to HIGH with SCL High
  SDA := SCL + 1
  outa[SCL]~~                       ' Drive SCL HIGH
  outa[SDA]~~                       '  then SDA HIGH
  dira[SCL]~                        ' Now let them float
  dira[SDA]~                        ' If pullups present, they'll stay HIGH

PUB Write(SCL, data) : ackbit | SDA
  SDA := SCL + 1
  ackbit := 0 
  data <<= 24
  repeat 8                          ' Output data to SDA
    outa[SDA] := (data <-= 1) & 1   
    outa[SCL]~~                     ' Toggle SCL from LOW to HIGH to LOW
    outa[SCL]~
  dira[SDA]~                        ' Set SDA to input for ACK/NAK
  outa[SCL]~~
  ackbit := ina[SDA]                ' Sample SDA when SCL is HIGH
  outa[SCL]~
  outa[SDA]~                        ' Leave SDA driven LOW
  dira[SDA]~~
'' PASM I2C Driver  Version 1.0
'' Copyright (c) 2010 Dave Hein
'' June 6, 2010
'' See end of file for terms of use

'' This is a PASM version of Mike Green's Basic I2C Driver.  The
'' low level I2C routines have been converted to PASM to increase the
'' I/O speed.  These routines use the same calling interface as
'' the Basic I2C Driver Version 1.1 in the OBEX, and should be fully
'' compatible with any existing code that uses the Basic I2C Driver.

'' Just like the Basic I2C Driver, the PASM I2C Driver assumes
'' that the SDA pin is one higher than the SCL pin.  It assumes that
'' neither the SDA nor the SCL pins have pullups, so drives both.

'' This object uses the Initialize method to start up a cog rather than using
'' the start method.  This is done to remain consistent with the Basic I2C
'' Driver routines.  Initialize must be called at the beginning of the program.
'' This loads the PASM code in a cog, and clocks the I2C bus to initialize
'' the devices on the bus.  Subsequent calls may be made to Initialize, and
'' it will not cause the cog to be stopped or reloaded.

'' The bus I/O speed is controlled by the constant DELAY_CYCLES.  This constant
'' is used in the delay routine.  The total delay consists of calling the delay
'' routine, performing a waitcnt of CNT + DELAY_CYCLES, and then returning
'' from the delay routine.  Therefore, the total delay will be about
'' 12 + DELAY_CYCLES.

'' The delay time represents the clock high time, and half the clock low time.
'' It is also used to determine the setup and hold times for the data bit for
'' read, write, start and stop operations.  DELAY_CYCLES is defined with a
'' value of 52, which gives a total delay of 64 cycles.  At 80 MHz, this is
'' 0.8 usecs, which is about one-third of a 400 KHz cycle time.  This value
'' should be modified to provide the optimal speed for a particular application.

'' Please see Mike Green's Basic I2C Driver object for more information on
'' the I2C routines, and on how EEPROMs are addressed
OBJ
  sys : "sysdefs"

CON
   ACK           = 0            ' I2C Acknowledge
   NAK           = 1            ' I2C No Acknowledge
   Xmit          = 0            ' I2C Direction Transmit
   Recv          = 1            ' I2C Direction Receive

   CMD_START     = 1            ' Issue a start bit
   CMD_STOP      = 2            ' Issue a stop bit
   CMD_READ      = 3            ' Transmit a byte to the I2C bus
   CMD_WRITE     = 4            ' Read a byte from the I2C bus
   CMD_INIT      = 5            ' Initialize the I2C bus
   CMD_READPAGE  = 6            ' Read one or more bytes in the page mode
   CMD_WRITEPAGE = 7            ' Write one or more bytes in the page mode
   
   DELAY_CYCLES  = 52           ' I2C Delay time.  Must be between 12 and 511

   i2c_cmd       = sys#i2c_cmd
   i2c_parm      = sys#i2c_parm
{
DAT
  cognum long 0
  cmdbuf long 0, 0

PUB Initialize(SCL)
'' Start cog if not started, and initialize the devices on the I2C bus
  'ser.dbprintf1(string("i2c initialize %d\n"), SCL)
  cmdbuf[1] := @SCL
  cmdbuf := CMD_INIT
  ifnot cognum
    cognum := cognew(@cmdloop, @cmdbuf) + 1
  repeat while cmdbuf
}

PUB Start(SCL)
'' Issue an I2C start command
  long[i2c_parm] := @SCL
  long[i2c_cmd] := CMD_START
  repeat while long[i2c_cmd]

PUB Stop(SCL)
'' Issue an I2C stop command
  long[i2c_parm] := @SCL
  long[i2c_cmd] := CMD_STOP
  repeat while long[i2c_cmd]

PUB Read(SCL, ackbit)
'' Read in i2c data, Data byte is output MSB first, SDA data line is
'' valid only while the SCL line is HIGH.  SCL and SDA left in LOW state.
  long[i2c_parm] := @SCL
  long[i2c_cmd] := CMD_READ
  repeat while long[i2c_cmd]
  result := long[i2c_parm]
  
PUB Write(SCL, data)
'' Write i2c data.  Data byte is output MSB first, SDA data line is valid
'' only while the SCL line is HIGH.  Data is always 8 bits (+ ACK/NAK).
'' SDA is assumed LOW and SCL and SDA are both left in the LOW state.
  long[i2c_parm] := @SCL
  long[i2c_cmd] := CMD_WRITE
  repeat while long[i2c_cmd]
  result := long[i2c_parm]
  
PUB ReadPage(SCL, devSel, addrReg, dataPtr, count) : ackbit
'' Read in a block of i2c data.  Device select code is devSel.  Device starting
'' address is addrReg.  Data address is at dataPtr.  Number of bytes is count.
'' The device select code is modified using the upper 3 bits of the 19 bit addrReg.
'' Return zero if no errors or the acknowledge bits if an error occurred.
  long[i2c_parm] := @SCL
  long[i2c_cmd] := CMD_READPAGE
  repeat while long[i2c_cmd]
  result := long[i2c_parm]
  
PUB WritePage(SCL, devSel, addrReg, dataPtr, count) : ackbit
'' Write out a block of i2c data.  Device select code is devSel.  Device starting
'' address is addrReg.  Data address is at dataPtr.  Number of bytes is count.
'' The device select code is modified using the upper 3 bits of the 19 bit addrReg.
'' Most devices have a page size of at least 32 bytes, some as large as 256 bytes.
'' Return zero if no errors or the acknowledge bits if an error occurred.  If
'' more than 31 bytes are transmitted, the sign bit is "sticky" and is the
'' logical "or" of the acknowledge bits of any bytes past the 31st.
  long[i2c_parm] := @SCL
  long[i2c_cmd] := CMD_WRITEPAGE
  repeat while long[i2c_cmd]
  result := long[i2c_parm]

PUB ReadByte(SCL, devSel, addrReg) : data
'' Read in a single byte of i2c data.  Device select code is devSel.  Device
'' starting address is addrReg.  The device select code is modified using the
'' upper 3 bits of the 19 bit addrReg.  This returns true if an error occurred.
   if ReadPage(SCL, devSel, addrReg, @data, 1)
      return -1

PUB ReadWord(SCL, devSel, addrReg) : data
'' Read in a single word of i2c data.  Device select code is devSel.  Device
'' starting address is addrReg.  The device select code is modified using the
'' upper 3 bits of the 19 bit addrReg.  This returns true if an error occurred.
   if ReadPage(SCL, devSel, addrReg, @data, 2)
      return -1

PUB ReadLong(SCL, devSel, addrReg) : data
'' Read in a single long of i2c data.  Device select code is devSel.  Device
'' starting address is addrReg.  The device select code is modified using the
'' upper 3 bits of the 19 bit addrReg.  This returns true if an error occurred.
'' Note that you can't distinguish between a return value of -1 and true error.
   if ReadPage(SCL, devSel, addrReg, @data, 4)
      return -1

PUB WriteByte(SCL, devSel, addrReg, data)
'' Write out a single byte of i2c data.  Device select code is devSel.  Device
'' starting address is addrReg.  The device select code is modified using the
'' upper 3 bits of the 19 bit addrReg.  This returns true if an error occurred.
   if WritePage(SCL, devSel, addrReg, @data, 1)
      return true
   return false

PUB WriteWord(SCL, devSel, addrReg, data)
'' Write out a single word of i2c data.  Device select code is devSel.  Device
'' starting address is addrReg.  The device select code is modified using the
'' upper 3 bits of the 19 bit addrReg.  This returns true if an error occurred.
'' Note that the word value may not span an EEPROM page boundary.
   if WritePage(SCL, devSel, addrReg, @data, 2)
      return true
   return false

PUB WriteLong(SCL, devSel, addrReg, data)
'' Write out a single long of i2c data.  Device select code is devSel.  Device
'' starting address is addrReg.  The device select code is modified using the
'' upper 3 bits of the 19 bit addrReg.  This returns true if an error occurred.
'' Note that the long word value may not span an EEPROM page boundary.
   if WritePage(SCL, devSel, addrReg, @data, 4)
      return true
   return false

PUB WriteWait(SCL, devSel, addrReg) : ackbit
'' Wait for a previous write to complete.  Device select code is devSel.  Device
'' starting address is addrReg.  The device will not respond if it is busy.
'' The device select code is modified using the upper 3 bits of the 18 bit addrReg.
'' This returns zero if no error occurred or one if the device didn't respond.
   devSel |= addrReg >> 15 & %1110
   Start(SCL)
   ackbit := Write(SCL, devSel | Xmit)
   Stop(SCL)
   return ackbit

{{
                            TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}
