'******************************************************************************
' Copyright (c) 2011, 2012, Dave Hein
' See end of file for terms of use.
'******************************************************************************
OBJ
  c : "clibsd"

CON
  GLOB_START  = 0
  GLOB_ACTIVE = 1
  GLOB_DONE   = 2

'************************************************************
' This routine returns a 1 if the passed string contains
' a wildcard character of either "*" or "?.  Otherwise,
' it will return a value of zero.
'************************************************************
PUB iswild(str) | char
  repeat while (char := byte[str++])
    if char == "*" or char == "?"
      return 1

'************************************************************
' This routine matches a string containing the wildcard
' characters of "*" and "?" to a string.
'************************************************************
PUB patmatch(pat, str) | pat1, str1
  repeat while byte[str] and byte[pat] <> "*"
    if byte[pat] <> byte[str] and byte[pat] <> "?"
      return 0
    pat++
    str++
  repeat while byte[str]
    if byte[pat] == "*"
      ifnot byte[++pat]
        return 1
      pat1 := pat
      str1 := str + 1
    elseif byte[pat] == byte[str] or byte[pat] == "?"
      pat++
      str++
    else
      pat := pat1
      str := str1++
  repeat while byte[pat] == "*"
    pat++
  return not byte[pat]

'************************************************************
' This routine will return matches to the string in pat
' if it contains wildcard characters.  It will return pat
' if it does not contain a wildcard character, or the
' directory path does not exist.  It returns a NULL pointer
' when there are no more items available.
'
' globptr points to a long that will contains the glob state
' information.  This long must be cleared prior to calling
' glob the first time for each pattern.
'
' str is used to hold the return value when a directory path
' is included in the pattern.  Wildcard characters cannot
' be used in the directory path, but can only be used in the
' file name located after the last "/".
'************************************************************
PUB glob(pat, globptr, str) | state, handle, pdirent, i, num, fname

  state  := byte[globptr][0]
  num    := byte[globptr][1]
  handle := word[globptr][1]

  if state == GLOB_START
    ifnot iswild(pat)
      byte[globptr][0] := GLOB_DONE
      return pat

    ' Locate the last "/" char
    c.strcpy(str, pat)
    i := num := 0
    repeat while byte[str][i]
      if byte[str][i] == "/"
        num := i
      i++

    ' Get the directory path
    if num
      byte[str][++num] := 0
      if iswild(str)
        byte[globptr][0] := GLOB_DONE
        return pat
    else
      byte[str] := 0
    byte[globptr][1] := num

    ' Open the directory
    handle := c.opendir(str)
    ifnot handle
      byte[globptr][0] := GLOB_DONE
      return pat
    byte[globptr][0] := state := GLOB_ACTIVE
    word[globptr][1] := handle

  if state == GLOB_ACTIVE
    repeat while (pdirent := c.readdir(handle))
      fname := pdirent + 8
      if strcomp(fname, string(".")) or strcomp(fname, string(".."))
        next
      ConvertToLower(pdirent+8)
      ifnot num
        if patmatch(pat, pdirent+8)
          return pdirent+8
      else
        if patmatch(pat+num, pdirent+8)
          c.strcpy(str+num, pdirent+8)
          return str
          
    c.closedir(handle)
    byte[globptr][0] := GLOB_DONE

PUB exit(globptr) | state, handle
  state  := byte[globptr][0]
  handle := word[globptr][1]
  if state == GLOB_ACTIVE
    c.closedir(handle)
  c.exit(1)

PUB ConvertToLower(ptr) | val
   repeat while val := byte[ptr]
     if (val => "A") and (val =< "Z")
       byte[ptr] := val + 32
     ptr++

{{
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
}}
