'******************************************************************************
' Copyright (c) 2011, 2012, Dave Hein
' See end of file for terms of use.
'******************************************************************************
CON
  _clkmode = xinput
  _xinfreq = $4e495053 ' SPIN

  ATTR_DIRECTORY = $10
  ATTR_READ_ONLY = $01

OBJ
  c : "clibsd"
  r : "resolve"

DAT
  buffer long 0[25]

PUB main(argc, argv) | fname1, fname2, attrib1, attrib2, ptr 

  c.enter

  if argc <> 3
    c.printf0(string("usage: mv fname1 fname2\n"))
    c.exit(0)

  fname1 := long[argv][1]
  fname2 := long[argv][2]

  r.resolve_path(fname1, @buffer)
  if strcomp(@buffer, string("/"))
    c.printf0(string("Can't move root directory\n"))
    c.exit(1)
  attrib1 := c.getmod(@buffer)

  r.resolve_path(fname2, @buffer)
  if strcomp(@buffer, string("/"))
    attrib2 := $30
  else
    attrib2 := c.getmod(@buffer)

  if attrib1 == -1
    c.printf1(string("%s does not exist\n"), fname1)
    c.exit(1)

  ' If fname1 is not a directory check if we should use copy
  ifnot attrib1 & ATTR_DIRECTORY
    ' Use copy if fname1 contains a "/"
    ptr := fname1
    repeat while byte[ptr] and byte[ptr] <> "/"
      ptr++
    if byte[ptr]
      c.exit(Copy(fname1, fname2, attrib1, attrib2))

    ' Use copy if fname2 contains a "/"
    ptr := fname2
    repeat while byte[ptr] and byte[ptr] <> "/"
      ptr++
    if byte[ptr]
      c.exit(Copy(fname1, fname2, attrib1, attrib2))

    ' Use copy if fname2 exists, and it is a directory
    if attrib2 <> -1 and (attrib2 & ATTR_DIRECTORY)
      c.exit(Copy(fname1, fname2, attrib1, attrib2))

  ' Otherwise, attempt to rename it
  result := c.rename(fname1, fname2)
  if result == -2
    c.printf1(string("%s already exists\n"), fname2)
  elseif result
    c.printf1(string("%s does not exist\n"), fname1)

  c.exit(0)

PUB Copy(fname1, fname2, attrib1, attrib2) | num, infile, outfile
  infile := c.fopen(fname1, string("r"))
  ifnot infile
    c.printf1(string("Could not open input file %s\n"), fname1)
    return 1
  if attrib1 & ATTR_DIRECTORY
    c.printf1(string("%s is a directory\n"), fname1)
    c.fclose(infile)
    return 1
  c.strcpy(@buffer, fname2)
  if attrib2 & ATTR_DIRECTORY
    num := strsize(@buffer)
    if num and byte[@buffer][num-1] <> "/"
      c.strcat(@buffer, string("/"))
    c.strcat(@buffer, GetFileName(fname1))
  outfile := c.fopen(@buffer, string("w"))
  ifnot outfile
    c.printf1(string("Could not open output file %s\n"), @buffer)
    c.fclose(infile)
    return 1
  c.chmod(@buffer, attrib1)
  repeat while (num := c.fread(@buffer, 1, 100, infile)) > 0
    c.fwrite(@buffer, 1, num, outfile)
  c.fclose(infile)
  c.fclose(outfile)
  if attrib1 & ATTR_READ_ONLY
    c.chmod(fname1, 0)
  c.remove(fname1)

PRI GetFileName(path) | len
  len := strsize(path)
  if len =< 0
    return path
  path += len

  repeat while len--
    if byte[--path] == "/"
      return path + 1

  return path

{{
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
}}
